//
//     
//
//   :       
// http://habrahabr.ru/blogs/algorithm/120051/
// http://habrahabr.ru/blogs/algorithm/122318/
//
//     O(n^2)
//        ;)
//
// http://robocraft.ru
//

#include <cv.h>
#include <highgui.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <string>

//   
// src -  
// wave_length -  
// base_light -  
IplImage* cvImageHologram(const IplImage* src, float wave_length, float base_light = 127.5 );
//    
IplImage* cvImageDehologram(const IplImage* src, float wave_length);

int main(int argc, char* argv[])
{
	IplImage *image = 0, *dst = 0, *dst2 = 0;

	char img_name[] = "Image0.jpg";

	//     
	char* image_filename = argc >= 2 ? argv[1] : img_name;

	//  
	image = cvLoadImage(image_filename, 1);

	printf("[i] image: %s\n", image_filename);
	if(!image){
		printf("[!] Error: cant load image: %s\n", image_filename);
		return -1;
	}

	//  
	cvNamedWindow( "image");
	cvShowImage( "image", image );

	//for(int i=1; i<1000; i++){

	//float wave_len = 2; //0.01*i;
	char* tmp_ptr=0; 
	float wave_len = argc >= 3 ? (float)strtod(argv[2], &tmp_ptr) : 0.0001;

	printf("%f\n", wave_len);

#if 1
	//  
	dst = cvImageHologram(image, wave_len);

	//    
	dst2 = cvImageDehologram(dst, wave_len);
#else
	//  
	dst = cvCreateImage(cvGetSize(image), IPL_DEPTH_8U, 1);
	cvConvertImage(image, dst, CV_BGR2GRAY);
	dst2 = cvImageDehologram(dst, wave_len);
#endif

	//  
	cvNamedWindow( "Hol");
	cvShowImage( "Hol", dst );

	cvNamedWindow( "deHol");
	cvShowImage( "deHol", dst2 );

#if 1
	//  
	std::string dst_file_name = image_filename;
	dst_file_name += "-hol.png";
	cvSaveImage(dst_file_name.c_str(), dst);
	dst_file_name = image_filename;
	dst_file_name += "-dehol.png";
	cvSaveImage(dst_file_name.c_str(), dst2);
#endif

#if 1
	//     
	int resW = image->width * 3;
	int resH = image->height * 3;
	IplImage* dst3 = cvCreateImage( cvSize(resW, resH), IPL_DEPTH_8U, 3);
	IplImage* dst4 = cvCreateImage( cvSize(resW, resH), IPL_DEPTH_8U, 1);

	//  
	cvNamedWindow( "obj");
	cvResize(image, dst3, CV_INTER_NN);
	cvShowImage( "obj", dst3 );
	cvNamedWindow( "Hologram");
	cvResize(dst, dst4, CV_INTER_NN);
	cvShowImage( "Hologram", dst4 );
	cvNamedWindow( "deHologram");
	cvResize(dst2, dst3, CV_INTER_NN);
	cvShowImage( "deHologram", dst3 );

	cvReleaseImage(&dst3);
	cvReleaseImage(&dst4);
#endif

	//   
	cvWaitKey(0);

	//  
	cvReleaseImage(&image);
	cvReleaseImage(&dst);
	cvReleaseImage(&dst2);

	//}

	//  
	cvDestroyAllWindows();
	return 0;
}

//   
// src -  
// wave_length -  
// base_light -   
IplImage* cvImageHologram(const IplImage* src, float wave_length, float base_light)
{
	if(!src){
		return 0;
	}

	float resize = 1;

	float W = wave_length; //   
	if(W<=0){
		return 0;
	}

	IplImage* dst = cvCreateImage(cvSize(src->width*resize, src->height*resize), IPL_DEPTH_32F, 1);
	if(!dst){
		return 0;
	}
	cvZero(dst);

	printf("[i] Start...\n");

	//      
	for( int y1=0; y1<dst->height; y1++ ) {
		float* ptr_dst = (float*) (dst->imageData + y1 * dst->widthStep);
		for( int x1=0; x1<dst->width; x1++ ) {
		
			//      - 
			//         

			//      
			for( int y0=0; y0<src->height; y0++ ) {
				uchar* ptr_src = (uchar*) (src->imageData + y0 * src->widthStep);
				for( int x0=0; x0<src->width; x0++ ) {

					//    (x0;y0)   (x1;y1)    z=1,  
					float D = sqrt( 1.0f + (x0*resize-x1)*(x0*resize-x1) + (y0*resize-y1)*(y0*resize-y1) );

					//     
					float Phase = D/W;

					//      :
					//      
					// (         )
					//    -     (x1;y1)

					// 3  
					ptr_dst[x1] +=  ptr_src[3*x0]*cos(Phase)   + // B - 
									ptr_src[3*x0+1]*cos(Phase) + // G - 
									ptr_src[3*x0+2]*cos(Phase);	 // R - 
				}
			}
		}
	}
	printf("[i] Done.\n");

	//   
	float max = 0;
	for( int y1=0; y1<dst->height; y1++ ) {
		float* ptr_dst = (float*) (dst->imageData + y1 * dst->widthStep);
		for( int x1=0; x1<dst->width; x1++ ) {
			if(ptr_dst[x1]>max){
				max = ptr_dst[x1];
			}
		}
	}

	printf("[i] max: %f \n", max);

	//  
	IplImage* temp = cvCreateImage(cvSize(src->width*resize, src->height*resize), src->depth, 1);

	//      
	for( int y1=0; y1<dst->height; y1++ ) {
		float* ptr_dst = (float*) (dst->imageData + y1 * dst->widthStep);
		uchar* ptr_temp = (uchar*) (temp->imageData + y1 * temp->widthStep);
		for( int x1=0; x1<dst->width; x1++ ) {
			float val = ptr_dst[x1];
			if(max!=0){
				//   
				val = abs( pow( (ptr_dst[x1]/max)*127.5 + base_light, 2 )/255 ); 
			}
			if(val>255.0){
				val = 255;
			}
			ptr_temp[x1] = (uchar)val;
		}
	}

	cvReleaseImage(&dst);

	return temp;
}

//    
// src -   
// wave_length -  
IplImage* cvImageDehologram(const IplImage* src, float wave_length)
{
	if(!src){
		return 0;
	}

	float resize = 1;

	IplImage* dst = cvCreateImage(cvSize(src->width*resize, src->height*resize), IPL_DEPTH_32F, 3);
	if(!dst){
		return 0;
	}
	cvZero(dst);

	float W = wave_length; //   
	if(W<=0){
		W=0.1;
	}

	printf("[i] Start...\n");

	//      
	for( int y1=0; y1<dst->height; y1++ ) {
		float* ptr_dst = (float*) (dst->imageData + y1 * dst->widthStep);
		for( int x1=0; x1<dst->width; x1++ ) {

			//      
			for( int y0=0; y0<src->height; y0++ ) {
				uchar* ptr_src = (uchar*) (src->imageData + y0 * src->widthStep);
				for( int x0=0; x0<src->width; x0++ ) {

					//    (x0;y0)   (x1;y1)    z=1,  
					float D = sqrt( 1.0f + (x0*resize-x1)*(x0*resize-x1) + (y0*resize-y1)*(y0*resize-y1) );

					//     
					float Phase = D/W;

					//      
					// (         )

					//   -     (x1;y1)

					float val = ptr_src[x0]*cos(Phase);
					// 3  
					ptr_dst[3*x1] += val;	// B - 
					ptr_dst[3*x1+1] += val; // G - 
					ptr_dst[3*x1+2] += val;	// R - 
				}
			}

			ptr_dst[3*x1] = pow(ptr_dst[3*x1], 2);
			ptr_dst[3*x1+1] = pow(ptr_dst[3*x1+1], 2);
			ptr_dst[3*x1+2] = pow(ptr_dst[3*x1+2], 2);
		}
	}
	printf("[i] Done.\n");

	//   
	float maxR = 0, maxG=0, maxB=0;
	for( int y1=0; y1<dst->height; y1++ ) {
		float* ptr_dst = (float*) (dst->imageData + y1 * dst->widthStep);
		for( int x1=0; x1<dst->width; x1++ ) {
			if(ptr_dst[3*x1]>maxB){
				maxB = ptr_dst[3*x1];
			}
			if(ptr_dst[3*x1+1]>maxG){
				maxG = ptr_dst[3*x1+1];
			}
			if(ptr_dst[3*x1+2]>maxR){
				maxR = ptr_dst[3*x1+2];
			}
		}
	}

	printf("[i] max: %f %f %f \n", maxR, maxG, maxB);

	IplImage* temp = cvCreateImage(cvSize(src->width*resize, src->height*resize), IPL_DEPTH_8U, 3);

	//     
	for( int y1=0; y1<dst->height; y1++ ) {
		float* ptr_dst = (float*) (dst->imageData + y1 * dst->widthStep);
		uchar* ptr_temp = (uchar*) (temp->imageData + y1 * temp->widthStep);
		for( int x1=0; x1<dst->width; x1++ ) {
			ptr_temp[3*x1] = ptr_dst[3*x1]*255/maxB;
			ptr_temp[3*x1+1] = ptr_dst[3*x1+1]*255/maxG;
			ptr_temp[3*x1+2] = ptr_dst[3*x1+2]*255/maxR;
		}
	}

	cvReleaseImage(&dst);

	return temp;
}
